//============================================================================
//
//                              Program/Module 
//                                   from
//                 Selfverifying Solvers for Dense Systems of 
//                     Linear Equations Realized in C-XSC
//
//                     Carlos Holbig and Walter Kraemer, 2003.
//
//       Program developed in C-XSC by Bernardo Frederes Kramer Alcalde,
//           Paulo Sergio Morandi Junior and Carlos Amaral Holbig
//
//============================================================================
//----------------------------------------------------------------------------
// File: ilss (implementation)
// Purpose: Compute approximations to the solution x of Ax = b and an
//   approximations of the inverse of A. In both case the system (m x n) can 
//   be square (m = n), over-determined (m > n) and under-determined (m < n).
// Global functions:
//   LSS(): General entry for linear system solver.
//   INV(): General entry for matrix inversion.
//----------------------------------------------------------------------------

#include "ilss.hpp"

using namespace cxsc;
using namespace std;

const real zerotest = 1e6;
const real delta    = 1e-15;
const real eps1     = 1e-15;
const real sqrt_01  = 0.31622777;
const real limit    = 1e20; 

static int m, n, dim;

static ivector null(ivector A)
{
  return (A = 0.0);
}

//----------------------------------------------------------------------------
// transposed imatrix
//----------------------------------------------------------------------------
static imatrix transp ( imatrix& A )
{                                            
  int      n;
  imatrix  res(Lb(A,2),Ub(A,2),Lb(A,1),Ub(A,1));

  for (n = Lb(A,1); n <= Ub(A,1); n++) Col(res,n) = Row(A,n);
  return res;
}


//----------------------------------------------------------------------------
// REL computes componentwise the maximum relative error of A w.r.t B. 
// if A[i] and B[i] do not have the same sign or if B[i] = 0, then
// rel. error = 0 for this component.
// A is always the new value of an iteration, B the old one.
//----------------------------------------------------------------------------
real REL(rvector A, rvector B)
{
  int i;
  real p,r,ai,bi;
 
  p=0;
  for(i=Lb(A);i<=Ub(A);i++) // A,B must have same index range
  {
    ai = A[i];
    bi = B[i];
    if( ai*bi <= 0.0 || zerotest*abs(ai) < abs(bi) ) r = 0.0;
      else r = abs( (ai-bi)/bi );
    if (r>p) p = r;
  }
  return p;
} // end REL


//----------------------------------------------------------------------------
// TOO_BAD = accuracy of A is far too bad
// note: 0 for false, 1 for true;
//----------------------------------------------------------------------------
bool TOO_BAD(ivector &A)
{
  int  i;
  bool bad;
 
  bad = false;
  for(i=Lb(A);i<=Ub(A);i++)
  {
    bad = bad || Inf(A[i]) < -limit && Sup(A[i]) > limit;
  }
  return bad;
} // end TOO_BAD


//----------------------------------------------------------------------------
// x1 is the new, x0 the old value of an iteration. If a component of x1
// has decreased by more than a factor of zerotest, then this component
// is set to 0. The same is done if the sign of a component has changed.
//----------------------------------------------------------------------------
void GUESS_ZEROES(rvector& x0, rvector& x1)
{
  int  i;
  real MAXX;
 
  MAXX = 0.0;
  for(i=Lb(x1);i<=Ub(x1);i++)
    if(abs(x1[i])>MAXX)
      for(i=Lb(x1);i<=Ub(x1);i++)
        if( x0[i]*x1[i] < 0.0 || zerotest*abs(x1[i]) < abs(x0[i]) 
	        && MAXX > zerotest*abs(x1[i]) )
          x1[i] = 0.0;
} // end GUESS_ZEROES


//----------------------------------------------------------------------------
// The result of Y is an enclosure of the solution
// errcode = 0: Y is enclosure of the solution
// errcode = 1: no enclosure obtained, bad condition (?)
// errcode = 2: no enclosure obtained, matrix A singular (?)
//
// FLAGS = 0: R1,R2,C have not yet been computed
// FLAGS = 1: only R1 has been computed
// FLAGS = 2: R1 and corresponding C have been computed
// FLAGS = 3: R1 and R2 are computedbut not the corresponding C
// FLAGS = 4: R1, R2 and the corresponding C have been computed
//----------------------------------------------------------------------------
void LSS(imatrix& A, ivector& b, ivector& Y, int& errcode, rmatrix& R1, 
         rmatrix& R2, imatrix& C, int& FLAGS)
{
  rmatrix       D( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) ), 
  	            R2temp( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) );
  rvector       x0( Lb(A,1),Ub(A,1) ), x1( Lb(A,1),Ub(A,1) ), 
	            y0( Lb(A,1),Ub(A,1) ), x1temp( Lb(A,1),Ub(A,1) ); 
  ivector       Y1( Lb(A,1),Ub(A,1) ), YA( Lb(A,1),Ub(A,1) ), 
	            Z( Lb(A,1),Ub(A,1) ),Y1temp( Lb(A,1),Ub(A,1) );
  int           i,j,k,err;
  bool          ready;
  real          p,bound,eps1;
  dotprecision  accu;
  idotprecision iaccu;
  interval      itemp;
  rmatrix       AM( Lb(A,1),Ub(A,1),Lb(A,2),Ub(A,2) );
  rvector       bm( Lb(b),Ub(b) );

  AM=mid(A); bm=mid(b);

  ready = false;
  
// begin USE_SINGLE_R - compute approximate inverse of A
  if (FLAGS<3)
  {
    err = 0;
    if (FLAGS<1)
    {
      R1 = AM;
      MINV( R1, err );
      if (err==0) FLAGS = 1;
    }
    if (err==0)
// floating point defect iteration: result is x1
    {
      bound = 100.0*sqrt_01;
      x1 = R1*bm;
      k = 0;
      do
      // iterate x = x + R*(b-Ax) 
      {
        k = k + 1;
        x0 = x1;
    
	    // x1 := #*(bm - AM*x0)
        for (i=Lb(x0);i<=Ub(x0);i++)
        {
          accu = bm[i];
          accumulate(accu,-AM[Row(i)],x0);
          x1[i] = rnd(accu);  
        }
		
	    // x1 := #*(x0 + R1*x1)
        for (i=Lb(x1);i<=Ub(x1);i++)
        {
          accu = x0[i];
          accumulate(accu,R1[Row(i)],x1);
          x1temp[i] = rnd(accu);  
        }
    
	    x1 = x1temp;
        p = REL(x1,x0);
        if(k>5) bound = bound*sqrt_01; 
      } while ((p<bound || k<=5) && p>=delta);
      
	  GUESS_ZEROES(x1,x0);
      
// Compute enclosure y0+y1 of the residuum b-A*x1 of the aproximation x1 
// and initialize Y1:=Z:= R1*(b-A*x1), C:= I-R1*A                                

      // y0 := #*(b-A*x1) = modified to y0:=MID(##(b - A*x1))
      for (i=Lb(x1);i<=Ub(x1);i++)
      {
        iaccu = b[i];
        accumulate(iaccu,-A[Row(i)],x1);
        rnd(iaccu,itemp);
        y0[i] = mid(itemp);
      } 
   
      // Y1 := ##(b-A*x1-y0)           
      for (i=Lb(x1);i<=Ub(x1);i++)
      {
        iaccu = b[i];
        accumulate(iaccu,-A[Row(i)],x1);
        iaccu = iaccu - y0[i];
        rnd(iaccu,Y1[i]);
      }  
   
      // Y1 := ##( R1*y0 + R1*Y1 );
      for (i=Lb(R1,1);i<=Ub(R1,1);i++)
      {
        iaccu = 0.0;
        accumulate(iaccu,R1[Row(i)],y0);
        accumulate(iaccu,R1[Row(i)],Y1);
        rnd(iaccu,Y1temp[i]);
      }
   
      Y1 = Y1temp;
      Z = Y1;
      
	  if (Z==null(Z))
      {
        Y = x1; // exact solution! (however, not necessarily unique!)
        errcode = 0;
        ready = true;
      }
        else
        {
          if (FLAGS<2)
          {
			// C := ##( ID(A) - R1*A );
            for (i=Lb(A,1);i<=Ub(A,1);i++)
              for (j=Lb(A,2);j<=Ub(A,2);j++)
              {
                iaccu = ( i == j) ? 1.0 : 0.0;
                accumulate(iaccu,-R1[i],A[Col(j)]);
                rnd(iaccu,C[i][j]);
              }

			FLAGS = 2;
          }

// interval iteration until inclusion is obtained (or max. iteration count)
          k = 0;
          eps1 = 0.1;
          do
          {
            if (k>=5) eps1 = 5*eps1;
            k = k+1;
            YA = Blow(Y1,eps1);
            Y1 = Z + C*YA;
            ready = in(Y1,YA);
          } while ( !ready && k<10 && !TOO_BAD(Y1) );

          // output of the result
          if (ready)
          {
            Y = x1 + Y1;
            errcode = 0;
          }    
        }
    }
      else ready = false;
  } // end USE_SINGLE_R
 
// if no success: try again with approximate inverse R = R1+R2 of double length
// USE_DOUBLE_R to try again with approximate inverse R = R1+R2 of double length 
  if (!ready)
  {
    err = 0;
    if (FLAGS<3)
    {
      R2 = R1*AM;
      MINV(R2,err);
      if (err==0)
      {
        FLAGS = 3;
        D = R2*R1;
    
	    // R2 := #* (R2*R1 - D);
        for (i=Lb(R1,1);i<=Ub(R1,1);i++)
          for (j=Lb(R1,2);j<=Ub(R1,2);j++)
          {
            accu = -D[i][j];
            accumulate(accu,R2[i],R1[Col(j)]);
            R2temp[i][j] = rnd(accu);
          }
        R2 = R2temp;
        R1 = D;
      }
    }

    if (err==0)
    // floating point defect iteration: result is x1+x0
    {
      bound = 100.0*sqrt_01;
   
      // x1 := #*( R1*bm + R2*bm );
      for (i=Lb(R1,1);i<=Ub(R1,1);i++)
      {  
        accu = 0.0;
        accumulate(accu,R1[Row(i)],bm);
        accumulate(accu,R2[Row(i)],bm);
        x1[i] = rnd(accu);
      }
   
      // x0 = #*( R1*bm + R2*bm - x1):
      for (i=Lb(R1,1);i<=Ub(R1,1);i++)
      {  
        accu = -x1[i];
        accumulate(accu,R1[Row(i)],bm);
        accumulate(accu,R2[Row(i)],bm);
        x0[i] = rnd(accu);
      }
   
      k = 0;
      do
      // iteration x = x + (R1+R2)*(b-Ax), x = x1 + x0)
      {
        k = k+1;
    
	    // y0 = #*(bm - AM*x1 - AM*x0)
        for (i=Lb(A,1);i<=Ub(A,1);i++)
        {  
          accu = bm[i];
          accumulate(accu,-AM[Row(i)],x1);
          accumulate(accu,-AM[Row(i)],x0);
          y0[i] = rnd(accu);
        }
    
	    // y0 := #*(x0 + R1*y0 + R2*y0);
        for (i=Lb(R1,1);i<=Ub(R1,1);i++)
        {  
          accu = x0[i];
          accumulate(accu,R1[Row(i)],y0);
          accumulate(accu,R2[Row(i)],y0);
          x1temp[i] = rnd(accu);
        }
        y0 = x1temp;
    
	    p = REL (x1+y0,x1+x0);
        y0 = x1 + y0;
    
	    // x0 := #*(x1 + x0 - y0)
        for (i=Lb(x1);i<=Lb(x1);i++)
        {
          accu = x1[i] + x0[i] - y0[i];
          x0[i] = rnd(accu); 
        }
   	    x1 = y0;
   
		if (k>5) bound = bound * sqrt_01;
      } while ( (p<bound || k<=5) && p>=delta );

// compute enclosure y0+Y1 of the residuum b-A*x1 of the approximation x1 and 
// initialize Y1:= Z:= (R1+R2)*(b-A*x1), C:= I-(R1+R2)*A   
   
      // y0 := #*(b-A*x1)  ====> y0:=MID(##(b - A*x1))
      for (i=Lb(x1);i<=Ub(x1);i++)
      {
        iaccu = b[i];
        accumulate(iaccu,-A[Row(i)],x1);
        rnd(iaccu, itemp);
        y0[i] = mid(itemp);  
      } 
   
      // Y1 := ##(b-A*x1-y0)           
      for (i=Lb(x1);i<=Ub(x1);i++)
      {
        iaccu = b[i];
        accumulate(iaccu,-A[Row(i)],x1);
        iaccu = iaccu - y0[i];
        rnd(iaccu,Y1[i]);
      }  
   
      // Y1 := ##(R1*y0 + R2*y0 + R1*Y1 + R2*Y1 );   
      for (i=Lb(R1,1);i<=Ub(R1,1);i++)
      {  
        accu = 0.0;
        accumulate(accu,R1[Row(i)],y0);
        accumulate(accu,R2[Row(i)],y0);
        iaccu = accu;
        accumulate(iaccu,R1[Row(i)],Y1);
        accumulate(iaccu,R2[Row(i)],Y1);
        rnd(iaccu,Y1temp[i]);
      }
      Y1 = Y1temp;
      Z = Y1;
   
      if (Z==null(Z))
      {
        Y = x1; // exact solution! (however, not necessarily unique!)
        errcode = 0;
        ready = true;
      }
        else
        {
          if (FLAGS<4)
          {
            // C:= ## (ID(A) - R1*A - R2*A);
            for (i=Lb(A,1);i<=Ub(A,1);i++)
              for (j=Lb(A,2);j<=Ub(A,2);j++)
              {
                iaccu = ( i == j) ? 1.0 : 0.0;
                accumulate(iaccu,-R1[i],A[Col(j)]);
                accumulate(iaccu,-R2[i],A[Col(j)]);
                rnd(iaccu,C[i][j]);
              }
     
	        FLAGS = 4;
          }
 
// interval iteration until inclusion is obtained (or max. iteration count)
          k = 0;
          eps1 = 0.1;
          do
          {
            if (k>=5) eps1 = 5*eps1;
            k = k+1;
            YA = Blow( Y1, eps1);
            Y1 = Z + C*YA;
            ready = in(Y1,YA);
          } while ( !ready && k<10 && !TOO_BAD(Y1) );

// output of the result
          if (ready)
          {
            Y = x1 + Y1;
            errcode = 0;
          }
            else errcode = 1;   
        }
    }
      else errcode = 2;
  } // end USE_DOUBLE_R
}  // end LSS


//----------------------------------------------------------------------------
// Linear system: square matrix
// The result y is an enclosure of the solution of Ax = b
//----------------------------------------------------------------------------
void SQUARE_LSS( imatrix& A, ivector& b, ivector& y, int& errcode) 
{
  rmatrix       R1(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW)),
                R2(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  imatrix       C(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  idotprecision Accu;
  int           FLAGS;

  if( dim == 1 )     // Treat trivial case separately
    if( in(0.0,A[Lb(A,ROW)][Lb(A,COL)]) ) errcode = 2;
      else
	  {
        y[Lb(y)] = b[Lb(b)] / A[Lb(A,ROW)][Lb(A,COL)];
        errcode = 0;
      }
    else
    {
      FLAGS = 0;
      LSS( A, b, y, errcode, R1, R2, C, FLAGS );
    }
} // end SQUARE_LSS


//----------------------------------------------------------------------------
// Linear system: over-determined case
// The result Y is an enclosure of the solution x of AH*A*x = AH*b,
// i.e. x is least squares solution of Ax = b.
//                                                 | A  -I | |x|   |b|
// From Ax = b we generate the (n+m)x(n+m)-system  |       | | | = | |
//                                                 | 0  AH | |y|   |0|
//
// Here, AH is the hermitian of A (transpose in the real case)
//----------------------------------------------------------------------------
void OVER_LSS( imatrix& A, ivector& b, ivector& y, int& errcode)
{
  imatrix BIG_A(1,dim,1,dim);
  ivector BIG_b(1,dim);
  ivector BIG_y(1,dim);
  int     i, j;

  BIG_A( 1,m, 1,n ) = A;
  
  // BIG_A( 1,m, n+1,n+m ) = -Id(m);
  for( i = 1; i <= m; i++)  
    for( j = n+1; j <= n+m; j++) 
      (j == i+2)? BIG_A[i][j]=-1 : BIG_A[i][j] = 0;  

  BIG_A( m+1,m+n, 1,n ) = 0.0;
  BIG_A( m+1,m+n, n+1,n+m ) = transp(A);
  BIG_b( 1,m ) = b;
  BIG_b( m+1,m+n ) = 0.0;
   
  SQUARE_LSS( BIG_A, BIG_b, BIG_y, errcode );  
 
  ivector_slice BIGy1n(BIG_y,1,n); //workaround;
  y=BIGy1n;                        //
                                   //replaces "y = BIG_y( 1,n );"
  //y = BIG_y( 1,n );              //g++-4.0.2 yields compiler error for
                                   //this
} // end OVER_LSS


//----------------------------------------------------------------------------
// Linear system: under-determined case
// The result Y is an enclosure of Y = AH*x with A*AH*x = b,
// i.e. y is solution of Ay = b with minimal Euklidian norm.
//                                                 |AH  -I | |x|   |0|
// From Ax = b we generate the (n+m)x(n+m)-system  |       | | | = | |
//                                                 | 0   A | |y|   |b|
//
// Here, AH is the hermitian of A (transpose in the real case)
//----------------------------------------------------------------------------
void UNDER_LSS( imatrix& A, ivector& b, ivector& y, int& errcode)
{
  imatrix BIG_A(1,dim,1,dim);
  ivector BIG_b(1,dim);
  ivector BIG_y(1,dim);
  int     i, j;

  BIG_A( 1,n, 1,m ) = transp(A);
    
  //BIG_A( 1,n, m+1,m+n ) = -Id(m);
  for( i = 1; i <= n; i++)
    for( j = m+1; j <= n+m; j++) 
      (j == i+2)? BIG_A[i][j]=-1 : BIG_A[i][j] = 0;   

  BIG_A( n+1,n+m, 1,m ) = 0.0;
  BIG_A( n+1,n+m, m+1,m+n ) = A;
  BIG_b( 1,n ) = 0.0;
  BIG_b( n+1,n+m ) = b;

  SQUARE_LSS( BIG_A, BIG_b, BIG_y, errcode );  

  ivector_slice BIGym1mn(BIG_y,m+1,m+n); //Workaround, mg2006
  y = BIGym1mn;                           //
  //y = BIG_y( m+1,m+n );
} // end UNDER_LSS


//----------------------------------------------------------------------------
// General entry for linear system solver: decides which case to treat
//----------------------------------------------------------------------------
void LSS( imatrix& A, ivector& b, ivector& y, int& errcode )
{
  errcode = 0;
  m = Ub(A,ROW) - Lb(A,ROW) + 1;
  n = Ub(A,COL) - Lb(A,COL) + 1;

  dim = m+n;
   
  if( m != Ub(b) - Lb(b) + 1 ) errcode = 3;  // b : wrong dimension
  if( n != Ub(y) - Lb(y) + 1 ) errcode = 4;  // y : wrong dimension

  if( errcode == 0 )
    if( m > n ) OVER_LSS( A, b, y, errcode ); // over-determined system
      else
        if( m < n ) UNDER_LSS( A, b, y, errcode ); // under-determined system
          else
		  {
            dim = n;
            SQUARE_LSS( A, b, y, errcode ); // square system
          }
} // end LSS


//----------------------------------------------------------------------------
// Inverse matrix: square case
// The result Y is an enclosure of the solution of AY = I
//----------------------------------------------------------------------------
void SQUARE_INV( imatrix& A, imatrix& Y, int& errcode) 
{
  rmatrix R1(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW)),
          R2(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  imatrix C(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  imatrix temp(Lb(A,ROW),Ub(A,ROW),Lb(A,ROW),Ub(A,ROW));
  ivector b(Lb(A,ROW),Ub(A,ROW));  
  ivector xTemp(Lb(A,ROW),Ub(A,ROW));
  int     i, err, FLAGS;

  if( dim == 1 )     // Treat trivial case separately
    if( in(0.0,A[Lb(A,ROW)][Lb(A,COL)]) ) errcode = 2;
      else
	  {
        Y[Lb(Y,ROW)][Lb(Y,COL)] = interval( 1.0 ) / A[Lb(A,ROW)][Lb(A,COL)];
        errcode = 0;
      }
    else
	{
      FLAGS = 0;
      err   = 0;
      b     = 0.0;
      for(i = Lb(A,ROW); i <= Ub(A,COL); i++)
	  {
        b[i] = 1.0;
        LSS( A,b,xTemp,errcode,R1,R2,C,FLAGS );
        Y[Lb(Y,ROW)+i-Lb(A,ROW)] = xTemp;
        if( errcode > err ) err = errcode;
        b[i] = 0.0;
      }
      errcode = err;
        
	  //Computing: Y = transp(Y);
      for( i = Lb(Y,ROW); i <= Ub(Y,COL); i++)
        Row(temp,i) = Col(Y,i); 
      Y = temp;
    }
} // end SQUARE_INV


//----------------------------------------------------------------------------
// Inverse matrix: over-determined case
// The result Y is an enclosure of the pseudo inverse A+ of A.
//                             -1
// If m > n then Y = A+ = (AH*A)  * AH is solution of:
//
// | A  -I | |Y|   |I|  ( A: mxn           )
// |       | | | = | |  ( X = AY - I: mxm  )
// | 0  AH | |X|   |0|  ( Y = A+    : nxm  )
//                      ( right hand side: )
//                      ( I : mxm, 0 : nxm )
//----------------------------------------------------------------------------
void OVER_INV( imatrix& A, imatrix& Y, int& errcode) 
{
  rmatrix R1(1,dim,1,dim), R2(1,dim,1,dim);
  imatrix BIG_A(1,dim,1,dim), C(1,dim,1,dim);
  ivector BIG_b(1,dim);
  ivector BIG_Y(1,dim);    
  int     i, j, err, FLAGS;

  BIG_A( 1,m, 1,n ) = A;
  
  // BIG_A( 1,m, n+1,n+m ) = -Id(m);
  for( i = 1; i <= m; i++)  
    for( j = n+1; j <= n+m; j++ ) 
      (j == i+2)? BIG_A[i][j]=-1 : BIG_A[i][j] = 0;  

  BIG_A( m+1,m+n, 1,n ) = 0.0;
  BIG_A( m+1,m+n, n+1,n+m ) = transp(A);
  BIG_b = 0.0;
  FLAGS = 0;
  err = 0;

  for( j = 1; j <= m; j++ )
  {
    BIG_b[j] = 1.0;
    LSS( BIG_A,BIG_b,BIG_Y,errcode,R1,R2,C,FLAGS );
    if( errcode > err ) err = errcode;
    BIG_b[j] = 0.0;
    ivector_slice BIGy1n(BIG_Y,1,n); //Workaround, mg2006
    Col(Y,Lb(Y,COL)+j-1) = BIGy1n;                 //
    //Col(Y,Lb(Y,COL)+j-1) = BIG_Y(1,n);           //(<-for this)
  }
  errcode = err;
} // end OVER_INV


//----------------------------------------------------------------------------
// Inverse matrix: under-determined case
// The result Y is an enclosure of the pseudo inverse A+ of A.
//                                -1
// If m < n then Y = A+ = AH*(A*AH)  is solution of:
//
// |AH  -I | |YH|   |I|  ( A: mxn           )
// |       | |  | = | |  ( X = AY - I: nxn  )
// | 0   A | |XH|   |0|  ( Y = A+    : nxm  )
//                       ( right hand side: )
//                       ( I : nxn, 0 : mxn )
//----------------------------------------------------------------------------
void UNDER_INV( imatrix& A, imatrix& Y, int& errcode) 
{
  rmatrix R1(1,dim,1,dim), R2(1,dim,1,dim);
  imatrix C(1,dim,1,dim), BIG_A(1,dim,1,dim);
  ivector BIG_b(1,dim);
  ivector BIG_Y(1,dim);    
  int     i, j, err, FLAGS;

  BIG_A( 1,n, 1,m ) = transp(A);
  
  //BIG_A( 1,n, m+1,m+n ) = -Id(m);
  for( i = 1; i <= n; i++)
    for( j = m+1; j <= n+m; j++) 
      (j == i+2)? BIG_A[i][j]=-1 : BIG_A[i][j] = 0;   

  BIG_A( n+1,n+m, 1,m )   = 0.0;
  BIG_A( n+1,n+m, m+1,m+n ) = A;
  BIG_b = 0.0;
  FLAGS = 0;
  err = 0;
  for( i = 1; i <= n; i++ )
  {
    BIG_b[i] = 1.0;
    LSS( BIG_A,BIG_b,BIG_Y,errcode,R1,R2,C,FLAGS );
    if( errcode > err ) err = errcode;
    BIG_b[i] = 0.0;
    ivector_slice BIGy1m(BIG_Y,1,m); // Workaround,mg2006
    Y[Lb(Y,ROW)+i-1] = BIGy1m;                     //
    //Y[Lb(Y,ROW)+i-1] = BIG_Y(1,m);               //<-(for this)
  }
  errcode = err;
} // end UNDER_INV


//----------------------------------------------------------------------------
// General entry for matrix inversion: decides which case to treat
//----------------------------------------------------------------------------
void INV( imatrix& A, imatrix& Y, int& errcode )
{
  errcode = 0;
  m = Ub(A,ROW) - Lb(A,ROW) + 1;
  n = Ub(A,COL) - Lb(A,COL) + 1;
  dim = m+n;
   
  if(n!= Ub(Y,ROW)-Lb(Y,ROW)+1 ) errcode = 3;  // Y : wrong number of rows
  if(m!= Ub(Y,COL)-Lb(Y,COL)+1 ) errcode = 4;  // Y : wrong number of columns

  if( errcode == 0 )
    if( m > n ) OVER_INV( A, Y, errcode ); // over-determined system
      else
        if( m < n ) UNDER_INV( A, Y, errcode ); // under-determined system
          else
		  {
            dim = n;
            SQUARE_INV( A, Y, errcode ); // square system
          }
} // end INV
